/*

 parport.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "parport.h"
#ifdef _PARPORT_H

#if defined _WIN32 || defined __CYGWIN__
#if defined INPOUT32 || defined PORTTALK
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef INPOUT32
#include "message.h"
#endif
#ifdef PORTTALK
#include <winioctl.h>
#define PORTTALK_TYPE 40000
#define IOCTL_IOPM_RESTRICT_ALL_ACCESS CTL_CODE(PORTTALK_TYPE, 0x900, METHOD_BUFFERED, FILE_ANY_ACCESS)
#define IOCTL_SET_IOPM CTL_CODE(PORTTALK_TYPE, 0x902, METHOD_BUFFERED, FILE_ANY_ACCESS)
#endif
#endif
#elif defined __FreeBSD__
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#elif defined __DJGPP__
#else /* Linux */
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#endif

#include "log.h"

#if defined _WIN32 || defined __CYGWIN__
#ifdef PORTTALK
static HANDLE porttalk = INVALID_HANDLE_VALUE;
#endif
#ifdef INPOUT32
static HINSTANCE hLib;
inpfuncPtr inp32 = NULL;
oupfuncPtr oup32 = NULL;
#endif
#elif defined __FreeBSD__
#ifdef PORTIO
static int iofd = -1;
#endif
#elif defined __DJGPP__
#else /* Linux */
#ifdef PORTIO
static unsigned int perm_port = 0;
#endif
#endif

#ifdef PARPORT
int parportfd = -1;
#endif

#ifdef INPOUT32
static void *lookup(const char *name) {
    void *addr = GetProcAddress(hLib, name);
    if (addr == NULL) { 
        DWORD error = GetLastError();
        log_printf("GetProcAddress for %s failed: %s(%u)", name, errormsg(error), (unsigned int)error); 
    }
    return addr;
}

int parport_libinit(void) {
    typedef BOOL (WINAPI *lpIsInpOutDriverOpen)(void);
    lpIsInpOutDriverOpen is_driver_open;
    if (hLib == NULL) {
        hLib = LoadLibrary("inpout32.dll");
        if (hLib == NULL) {
            DWORD error = GetLastError();
            log_printf("LoadLibrary of inpout32.dll failed: %s(%u)", errormsg(error), (unsigned int)error); 
            return 0; 
        }
    }
    is_driver_open = (lpIsInpOutDriverOpen)lookup("IsInpOutDriverOpen");
    if (is_driver_open == NULL) return 0;
    if (!is_driver_open()) {
        log_print("InpOut32 driver is not loaded");
        return 0;
    }
    inp32 = (inpfuncPtr)lookup("Inp32");
    if (inp32 == NULL) return 0;
    oup32 = (oupfuncPtr)lookup("Out32");
    if (oup32 == NULL) return 0;
    return 1;
}

void parport_libdeinit(void) {
    if (hLib != NULL) {
        FreeLibrary(hLib);
        hLib = NULL;
    }
    inp32 = NULL;
    oup32 = NULL;
}
#endif

int parport_init(const char *device, unsigned int port) {
#if defined _WIN32 || defined __CYGWIN__
    (void)device;
#ifdef INPOUT32
    if (inp32 != NULL && oup32 != NULL) {
        return 0;
    }
#endif
#ifdef PORTTALK
    if (porttalk == INVALID_HANDLE_VALUE) {
        porttalk = CreateFile("\\\\.\\PortTalk", GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
        if (porttalk == INVALID_HANDLE_VALUE) {
            log_print("PortTalk driver not found");
            return 1;
        }
    }
    if (porttalk != INVALID_HANDLE_VALUE) {
        DWORD ret;
        DeviceIoControl(porttalk, IOCTL_IOPM_RESTRICT_ALL_ACCESS, NULL, 0, NULL, 0, &ret, NULL);
        if (DeviceIoControl(porttalk, IOCTL_SET_IOPM, &port, 3, NULL, 0, &ret, NULL) != 0) {
            log_print("PortTalk IOCTL_SET_IOPM failed");
            return 1;
        }
    }
#else
    (void)port;
#endif
#elif defined __FreeBSD__
#ifdef PARPORT
    if (device != NULL && parportfd == -1) {
        unsigned char status;
        parportfd = open(device, O_RDWR);
        if (parportfd == -1) {
            log_printf("Error opening %s ppi device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }
        if (ioctl(parportfd, PPIGSTATUS, &status) == -1) {
            log_printf("Error reading status of %s ppi device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }
    } else 
#endif
    {
        (void)port;
#ifdef PORTIO
        (void)device;
        if (iofd == -1) {
            iofd = open("/dev/io", O_RDWR);
            if (iofd == -1) {
                log_printf("Error opening /dev/io: %s(%d)", strerror(errno), errno);
                return 1;
            }
        }
#else
        log_print("Direct port access is unavailable, use a ppi device");
        return 1;
#endif
    }
#elif defined __DJGPP__
    (void)device;
    (void)port;
#else /* Linux */
#ifdef PARPORT
    if (device != NULL && parportfd == -1) {
        int mode;
        parportfd = open(device, O_RDWR);
        if (parportfd == -1) {
            log_printf("Error opening %s parport device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }
        if (ioctl(parportfd, PPEXCL) || ioctl(parportfd, PPCLAIM)) {
            close(parportfd);
            parportfd = -1;
            log_printf("Cannot access %s parport device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }
        mode = IEEE1284_MODE_ECP;
        if (ioctl(parportfd, PPSETMODE, &mode)) {
            close(parportfd);
            parportfd = -1;
            log_printf("Cannot set byte mode on %s parport device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }

        mode = 0;
        if (ioctl(parportfd, PPDATADIR, &mode)) {
            close(parportfd);
            parportfd = -1;
            log_printf("Cannot set direction on %s parport device: %s(%d)", device, strerror(errno), errno);
            return 1;
        }
    } else
#endif
    {
#ifdef PORTIO
        if (perm_port == 0) {
            if (ioperm(port, 3, 1)) {
                log_printf("Cannot allocate I/O ports %x to %x: %s(%d)", port, port + 2, strerror(errno), errno);
                return 1;
            } else perm_port = port;
        }
#else
        (void)port;
        log_print("Direct port access is unavailable, use a parport device");
        return 1;
#endif
    }
#endif
    return 0;
}

void parport_deinit(void) {
#if defined _WIN32 || defined __CYGWIN__
#ifdef PORTTALK
    if (porttalk != INVALID_HANDLE_VALUE) {
        DWORD ret;
        DeviceIoControl(porttalk, IOCTL_IOPM_RESTRICT_ALL_ACCESS, NULL, 0, NULL, 0, &ret, NULL);
        CloseHandle(porttalk);
        porttalk = INVALID_HANDLE_VALUE;
    }
#endif
#elif defined __FreeBSD__
#ifdef PARPORT
    if (parportfd != -1) {
        close(parportfd);
        parportfd = -1;
    } else
#endif
    {
#ifdef PORTIO
        if (iofd != -1) {
            close(iofd);
            iofd = -1;
        }
#endif
    }
#elif defined __DJGPP__
#else /* Linux */
#ifdef PARPORT
    if (parportfd != -1) {
        ioctl(parportfd, PPRELEASE);
        close(parportfd);
        parportfd = -1;
    } else
#endif
    {
#ifdef PORTIO
        if (perm_port != 0) {
            ioperm(perm_port, 3, 0);
            perm_port = 0;
        }
#endif
    }
#endif
}
#endif
