/*

 partition.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "partition.h"
#include <string.h>
#include <stdlib.h>

static partition_t work_partition;
static char null_path;

static Partition *partition(partition_t n) {
    static Partition partitions[256];
    if (n == 0) n = work_partition;
    if (n < 1 || n > 255) return NULL;
    return &partitions[n];
}

void partition_create(partition_t n, const Petscii *name, const char *root) {
    Partition *p = partition(n);
    unsigned int i;

    if (p == NULL || p->path != NULL) return;
    for (i = 0; name[i] != '\0' && i < sizeof(p->name) - 1; i++) {
        p->name[i] = name[i];
    }
    p->name[i] = 0;
    p->root = strlen(root);
    if (p->root == 0) {
        p->path = &null_path;
    } else {
        size_t len = p->root + 1;
        p->path = (char *)malloc(len);
        if (p->path == NULL) exit(EXIT_FAILURE);
        memcpy(p->path, root, len);
    }
}

static void partition_delete(partition_t n) {
    Partition *p = partition(n);

    if (p == NULL) return;
    if (p->path != &null_path) {
        free(p->path);
        p->path = NULL;
    }
}

int partition_select(partition_t n) {
    Partition *p = partition(n);

    if (p == NULL || p->path == NULL) return 1;
    work_partition = n;
    return 0;
}

const Partition *partition_get(partition_t n) {
    Partition *p = partition(n);
    if (p == NULL || p->path == NULL) return NULL;
    return p;
}

partition_t partition_get_current(void) {
    return work_partition;
}

void partition_set_path(partition_t n, const char *path) {
    Partition *p = partition(n);
    size_t len;

    if (p == NULL || p->path == NULL) return;
    if (path[0] == 0) {
        if (p->path != &null_path) {
            free(p->path);
            p->path = &null_path;
        }
        return;
    }
    len = strlen(path) + 1 + p->root;
    if (len > strlen(p->path) + 1) {
        char *nm = (char *)(p->path == &null_path ? malloc(len) : realloc(p->path, len));
        if (nm == NULL) return;
        p->path = nm;
    }
    if (len > p->root) {
        memcpy(p->path + p->root, path + p->root, len - p->root);
    } else {
        p->path[p->root] = 0;
    }
}

void partition_destroy(void) {
    unsigned int i;
    for (i = 1; i < 256; i++) {
        partition_delete(i);
    }
}
