/*

 path.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "path.h"
#include "partition.h"
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include "log.h"
#include "ideservd.h"
#include "directory.h"
#include "petscii.h"
#include "wchar.h"
#include "utf8.h"

void path_reset(Path *path) {
    path->data = NULL;
    path->len = 0;
    path->max = 0;
    path->root = 0;
}

const char *path_new(Path *path, const char *s, size_t root) {
    size_t len = strlen(s) + 1;
    char *data = (char *)malloc(len);
    if (data != NULL) {
        path->data = data;
        path->len = len - 1;
        path->max = len;
        path->root = root;
        memcpy(data, s, len);
    } else path_reset(path);
    return data;
}

void path_free(Path *path) {
    free(path->data);
}

const char *path_append(Path *path, const char *s) {
    char *data = path->data;
    if (s[0] != '\0') {
        unsigned int separator = (path->len != 0 && path->data[path->len - 1] != '/') ? 1 : 0;
        size_t len = strlen(s) + 1;
        size_t len2 = len + path->len + separator;
        if (len2 > path->max) {
            data = (char *)realloc(path->data, len2);
            if (data == NULL) return NULL;
            path->data = data;
            path->max = len2;
        }
        if (separator != 0) data[path->len++] = '/';
        memcpy(data + path->len, s, len);
        path->len += len - 1;
    }
    return data;
}

void path_root(Path *path) {
    if (path->data == NULL) return;
    path->data[path->root] = '\0';
    path->len = path->root;
}

const char *path_parent(Path *path) {
    if (path->data != NULL) {
        while (path->len > path->root && path->data[path->len - 1] != '/') {
            path->len--;
        }
        if (path->len > path->root) {
            path->len--;
        }
        path->data[path->len] = '\0';
    }
    return path->data;
}

int invalid_c64name(Petscii *lname) {
    int f;
    for (f = 0; lname[f] != 0; f++) {
        if (lname[f] == '*' || lname[f] == '?' || lname[f] == ':' || lname[f] == '=') {
            return 1;
        }
    }
    return 0;
}

void convertfilename(const Petscii *s, char typek, Petscii *name, Petscii *type, Petscii *mode) {
    int b, i, j;
    Petscii type2[4], mode2;
    int comma;

    i = comma = mode2 = 0; j = -1;

    for (b = 0; s[b] != 0; b++) {
        if (s[b] == ':') {
            i = comma = mode2 = 0; j = -1; continue;
        }
        if (s[b] == '/') {
            if (typek == 0) break;
        }

        switch (comma) {
        case 0:
            if (s[b] == typek) {
                comma++; j = 0;
                break;
            }
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            if (i < 16) {
                name[i++] = s[b];
            }
            break;
        case 1:
            if (s[b] == ',') {
                comma++;
                break;
            }
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            if (j < 3) {
                type2[j++] = s[b];
            }
            break;
        case 2:
            if (s[b] == ',' || s[b] == '=') {
                comma = 3;
                break;
            }
            mode2 = s[b];
            comma++; break;
        default:
            break;
        }
    }
    name[i] = 0;
    if (j == 1) {
        if (mode2 == 0) {
            switch (type2[0]) {
            case 'W':
            case 'R':
            case 'A':
            case 'M':
                mode2 = type2[0];
                if (mode2 == 'W' && type != NULL && type[0] == '*') {
                    type2[0] = 'S';
                    break;
                }
                j = -1;
                break;
            }
        }
    }
    if (j >= 0 && type != NULL) {
        if (j == 1) {
            switch (type2[0]) {
            case 'S': memcpy(type2, "SEQ", 4); j = 3; break;
            case 'P': memcpy(type2, "PRG", 4); j = 3; break;
            case 'U': memcpy(type2, "USR", 4); j = 3; break;
            case 'B': memcpy(type2, "DIR", 4); j = 3; break;
            case 'J': memcpy(type2, "LNK", 4); j = 3; break;
            default: type2[0] = ':'; break;
            }
        }
        memcpy(type, type2, (unsigned int)j);
        type[j] = 0;
    }
    if (mode2 && mode) {
        *mode = mode2;
    }
}

int matchname(Petscii *a, Petscii *b) {
    while (*a == *b || (*b == '?' && *a != 0)) {
        if (*a == 0) {
            return 1;
        }
        a++; b++;
    }

    if (*b == '*') {
        b++;
        if (*b == 0) {
            return 1;
        }
        while (*a != 0) {
            if (matchname(a, b)) {
                return 1;
            }
            a++;
        }
    }

    return 0;
}

const Petscii *resolv_part(const Petscii *s, Path *outpath, partition_t *outpart) {
    partition_t partition = 0;
    int fel = 0;
    if (strchr((const char *)s, ':')) {
        if (s[0] == '@') fel = 1;
        while ((s[fel] ^ 0x30) < 10) partition = partition * 10 + (s[fel++] ^ 0x30);
        while (s[fel] && s[fel] != '/' && s[fel] != ':') {
            fel++;
        }
        fel++;
    }
    *outpart = partition;
    {
        const Partition *p = partition_get(partition);
        if (p == NULL) {
            path_reset(outpath);
            return NULL;
        }
        if (path_new(outpath, p->path, p->root) == NULL) {
            log_print("Path: Out of memory");
            return NULL;
        }
        if (s[fel] == '/') path_root(outpath);
    }
    return s + fel;
}

const Petscii *resolv_path(const Petscii *s, Path *outpath, Nameconversion nameconversion) {
    int fel = 0, fel2 = 0;
    for (;;) {
        Petscii name[17];
        Directory_entry direntry;
        Directory *directory;
        int found = 0;

        if (s[fel] == '/' || s[fel] == ':') {
            fel++; continue;
        }

        fel2 = fel;
        if (s[fel] != 0) {
            while (s[fel2] != 0 && s[fel2] != '/' && s[fel2] != ':') {
                fel2++;
            }
        }
        if (s[fel2] == 0) {
            return s + fel;
        }

        if (s[fel] == '.') {
            if (fel2 == fel + 1) {
                fel += 2; continue;
            }
            if (s[fel + 1] == '.' && fel2 == fel + 2) {
                path_parent(outpath);
                fel += 3; continue;
            }
        }

        directory = directory_open(outpath, nameconversion, 0);
        if (directory == NULL) {
            errnotochannel15(errno);
            log_printf("Couldn't open the directory \"%s\": %s(%d)", outpath->data, strerror(errno), errno);
            return NULL;
        }
        convertfilename(s + fel, 0, name, NULL, NULL);

        while (directory_read(directory, &direntry))
        {
            if ((direntry.attrib & (A_DIR | A_CLOSED)) != (A_CLOSED | A_DIR)) {
                continue;
            }

            if (!matchname(direntry.name, name)) {
                continue;
            }

            if (path_append(outpath, directory_filename(directory)) == NULL) {
                log_print("Path: Out of memory");
                break;
            }
            found = 1;
            break;
        }
        directory_close(directory);
        if (!found) {
            Petscii lname[MAX_C64NAME_LEN];
            convertc64name(lname, name, NULL, nameconversion);
            if (path_appendc64(outpath, lname) == NULL) log_print("Path: Out of memory");
            return NULL;
        }
        fel = fel2;
    }
}

static int badfilename(const Petscii *name) {
    if (name[0] == '.') {
        if (name[1] == 0) return 1;
        if (name[1] == '.' && name[2] == 0) return 1;
    }
    return 0;
}

void convertc64name(Petscii lname[MAX_C64NAME_LEN], const Petscii *name, const Petscii *type, Nameconversion nameconversion) {
    unsigned int i, f2;
    int doext;
    Petscii dot;
    for (f2 = 0; name[f2] != 0 && f2 < 16; f2++) {
        lname[f2] = name[f2];
    }
    lname[f2] = 0;
    if (type == NULL) {
        return;
    }

    dot = nameconversion == NC_FORCEDOT ? '.' : ',';
    if (nameconversion == NC_IGNOREDOT) {
        doext = memcmp(type, "PRG", 3) != 0;
    } else {
        doext = type[0] != 0 || strchr((const char *)((lname[0] == '.') ? (lname + 1) : lname), '.') != NULL || badfilename(lname);
    }

    for (;;) {
        unsigned int f = f2;
        if (doext) {
            lname[f2++] = dot;
            for (i = 0; type[i] != 0 && i < 3; i++) {
                if (i == 1 && type[1] == 0xa0 && type[2] == 0xa0) break;
                lname[f2++] = type[i];
            }
            lname[f2] = 0;
        }
        if (!badfilename(lname)) break;
        doext = 1;
        f2 = f;
        dot = ',';
    }
}

#ifdef _WIN32
static unsigned int petscii_to_utf8(Petscii c, char *d) {
    unsigned int w = petscii_to_unicode(c);
    if (w < 0x80) {
        *d = (char)w;
        return 1;
    }
    return utf8out(w, d);
}

const char *path_appendc64(Path *path, const Petscii *name) {
    char lname[6 * MAX_C64NAME_LEN];
    unsigned int i, f2 = 0;
    for (i = 0; name[i] != 0; i++) {
        f2 += petscii_to_utf8(name[i], lname + f2);
    }
    lname[f2] = 0;
    return path_append(path, lname);
}
#else
static size_t c64toascii(Petscii c, char *d, mbstate_t *ps) {
    ssize_t l;
    unsigned int w = petscii_to_unicode(c);
#if WCHAR_MAX <= 0xFFFFu
    if (w >= 0x10000) {
        l = wcrtomb(d, (wchar_t)(0xd800 + ((w - 0x10000) >> 10)), ps);
        if (l >= 0) {
            ssize_t l2 = wcrtomb(d + l, (wchar_t)(0xdc00 + (w & 0x3ff)), ps);
            l = l2 < 0 ? l2 : (l + l2);
        }
    } else
#endif
    if (w > '~') {
        l = (ssize_t)wcrtomb(d, (wchar_t)w, ps);
    } else {
        l = 1; *d = (char)w;
    }
    if (l >= 0) return (size_t)l;
#if defined _WIN32 || defined __CYGWIN__ || defined __DJGPP__
    l = sprintf(d, "?");
#else
    l = sprintf(d, w <= 0xffff ? "\\u%04x" : "\\U%08x", (unsigned int)w);
#endif
    return (l >= 0) ? (size_t)l : 0;
}

const char *path_appendc64(Path *path, const Petscii *name) {
    char lname[10 * MAX_C64NAME_LEN];
    mbstate_t ps;
    unsigned int i, f2 = 0;
    memset(&ps, 0, sizeof ps);
    for (i = 0; name[i] != 0; i++) {
        f2 += c64toascii(name[i], lname + f2, &ps);
    }
    lname[f2] = 0;
    return path_append(path, lname);
}
#endif

