/*

 utf8.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "utf8.h"
#ifdef _WIN32
#include <stdlib.h>
#include <direct.h>
#include <errno.h>
#include <sys/stat.h>
#include <string.h>
#include "wchar.h"

unsigned int utf8in(const char *c, unsigned int *out) {
    unsigned int i, j;
    unsigned int ch = (unsigned char)c[0];

    if (ch < 0xe0) {
        *out = (ch << 6) ^ (unsigned char)c[1] ^ 0x3080;
        return 2;
    }
    if (ch < 0xf0) {
        ch ^= 0xe0;i = 3;
    } else if (ch < 0xf8) {
        ch ^= 0xf0;i = 4;
    } else if (ch < 0xfc) {
        ch ^= 0xf8;i = 5;
    } else {
        ch ^= 0xfc;i = 6;
    }

    for (j = 1;j < i; j++) {
        ch = (ch << 6) ^ (unsigned char)c[j] ^ 0x80;
    }
    *out = ch;
    return i;
}

unsigned int utf8out(unsigned int i, char *c) {
    if (i < 0x800) {
        c[0] = (char)(0xc0 | (i >> 6));
        c[1] = (char)(0x80 | (i & 0x3f));
        return 2;
    }
    if (i < 0x10000) {
        c[0] = (char)(0xe0 | (i >> 12));
        c[1] = (char)(0x80 | ((i >> 6) & 0x3f));
        c[2] = (char)(0x80 | (i & 0x3f));
        return 3;
    }
    if (i < 0x200000) {
        c[0] = (char)(0xf0 | (i >> 18));
        c[1] = (char)(0x80 | ((i >> 12) & 0x3f));
        c[2] = (char)(0x80 | ((i >> 6) & 0x3f));
        c[3] = (char)(0x80 | (i & 0x3f));
        return 4;
    }
    if (i < 0x4000000) {
        c[0] = (char)(0xf8 | (i >> 24));
        c[1] = (char)(0x80 | ((i >> 18) & 0x3f));
        c[2] = (char)(0x80 | ((i >> 12) & 0x3f));
        c[3] = (char)(0x80 | ((i >> 6) & 0x3f));
        c[4] = (char)(0x80 | (i & 0x3f));
        return 5;
    }
    if ((i & ~(unsigned int)0x7fffffff) != 0) return 0;
    c[0] = (char)(0xfc | (i >> 30));
    c[1] = (char)(0x80 | ((i >> 24) & 0x3f));
    c[2] = (char)(0x80 | ((i >> 18) & 0x3f));
    c[3] = (char)(0x80 | ((i >> 12) & 0x3f));
    c[4] = (char)(0x80 | ((i >> 6) & 0x3f));
    c[5] = (char)(0x80 | (i & 0x3f));
    return 6;
}

static wchar_t *utf8_to_wchar_common(const char *name, wchar_t *wname, size_t len, size_t j) {
    enum { REPLACEMENT_CHARACTER = 0xfffd };
    unsigned int ch;
    size_t i = 0;
    while (name[i] != 0) {
        ch = (unsigned char)name[i];
        if ((ch & 0x80) != 0) {
            i += utf8in(name + i, &ch);
            if (ch == 0) ch = REPLACEMENT_CHARACTER;
        } else i++;
        if (j + 3 > len) {
            wchar_t *d;
            len += 64;
            if (len < 64) goto failed;
            d = (wchar_t *)realloc(wname, len * sizeof *wname);
            if (d == NULL) goto failed;
            wname = d;
        }
        if (ch < 0x10000) {
        } else if (ch < 0x110000) {
            wname[j++] = (wchar_t)((ch >> 10) + 0xd7c0);
            ch = (ch & 0x3ff) | 0xdc00;
        } else ch = REPLACEMENT_CHARACTER;
        wname[j++] = (wchar_t)ch;
    }
    wname[j] = 0;
    return wname;
failed:
    free(wname);
    return NULL;
}

wchar_t *utf8_to_wchar(const char *name) {
    wchar_t *wname;
    size_t len = strlen(name) + 2;
    wname = (wchar_t *)malloc(len * sizeof *wname);
    if (wname == NULL) return NULL;
    return utf8_to_wchar_common(name, wname, len, 0);
}

char *wchar_to_utf8(const wchar_t *wname) {
    unsigned int ch;
    size_t i = 0, j = 0, len = wcslen(wname) + 10;
    char *name = (char *)malloc(len);
    if (name == NULL) return NULL;
    while (wname[i] != 0) {
        ch = (unsigned int)wname[i++];
        if (ch >= 0xd800 && ch < 0xdc00 && wname[i] >= 0xdc00 && wname[i] < 0xe000) {
            ch = wname[i++] ^ 0x360dc00 ^ (ch << 10);
            ch += 0x10000;
        }
        if (j + 6 >= len) {
            char *d;
            len += 64;
            if (len < 64) goto failed;
            d = (char *)realloc(name, len);
            if (d == NULL) goto failed;
            name = d;
        }
        if (ch < 0x80) name[j++] = (char)ch; else j += utf8out(ch, name + j);
    }
    name[j] = 0;
    return name;
failed:
    free(name);
    return NULL;
}

static char *utf8_to_char_common(const char *s, char *d, size_t len, size_t j) {
    mbstate_t ps;
    size_t i = 0;
    memset(&ps, 0, sizeof ps);
    while (s[i] != 0) {
        ssize_t l;
        unsigned int ch = (unsigned char)s[i];
        if ((ch & 0x80) != 0) i += utf8in(s + i, &ch); else i++;
        if (j + 10 >= len) {
            char *d2;
            len += 64;
            if (len < 64) goto failed;
            d2 = (char *)realloc(d, len);
            if (d2 == NULL) goto failed;
            d = d2;
        }
#if WCHAR_MAX <= 0xFFFFu
        if (ch >= 0x10000) {
            l = wcrtomb(d + j, (wchar_t)(0xd800 + ((ch - 0x10000) >> 10)), &ps);
            if (l >= 0) {
                ssize_t l2 = wcrtomb(d + j + l, (wchar_t)(0xdc00 + (ch & 0x3ff)), &ps);
                l = l2 < 0 ? l2 : (l + l2);
            }
        } else
#endif
            if (ch > '~') {
                l = (ssize_t)wcrtomb(d + j, (wchar_t)ch, &ps);
            } else {
                l = 1; d[j] = (char)ch;
            }
        if (l < 0) {
#if defined WIN32 || defined __DJGPP__
            l = 1; strcpy(d + j, "?");
#else
            l = sprintf(d + j, ch <= 0xffff ? "\\u%04x" : "\\U%08x", (unsigned int)ch);
#endif
            if (l < 0) goto failed;
        }
        j += (size_t)l;
    }
    d[j] = 0;
    return d;
failed:
    free(d);
    return NULL;
}

char *utf8_to_char(const char *s) {
    size_t len = strlen(s) + 10;
    char *d = (char *)malloc(len);
    if (d == NULL) return NULL;
    return utf8_to_char_common(s, d, len, 0);
}

char *char_to_utf8(const char *name) {
    mbstate_t ps;
    unsigned int ch;
    size_t i = 0, j = 0, len = strlen(name) + 10;
    char *utf8 = (char *)malloc(len);
    if (utf8 == NULL) return NULL;
    memset(&ps, 0, sizeof ps);
    while (name[i] != 0) {
        const char *s = name + i;
        ssize_t l;
        if ((unsigned char)*s <= '~') {
#if defined WIN32 || defined __DJGPP__
            ch = *s; l = 1;
#else
            if (*s == '\\' && (s[1] == 'u' || s[1] == 'U')) {
                int j;
                ch = 0; l = 2;
                for (j = (s[1] == 'u') ? 12 : 28; j >= 0; j -= 4, l++) {
                    if (s[l] >= '0' && s[l] <= '9') ch |= (unsigned int)(s[l] - '0') << j;
                    else if (s[l] >= 'a' && s[l] <= 'f') ch |= (unsigned int)(s[l] - 'a' + 10) << j;
                    else {
                        ch = '\\'; l = 1;
                        break;
                    }
                }
            } else {
                ch = *(unsigned char *)s; l = 1;
            }
#endif
        } else {
#if WCHAR_MAX <= 0xFFFFu
            wchar_t w2, w3;
            l = mbrtowc(&w2, s, MB_CUR_MAX, &ps);
            if (l < 0) goto failed;
            if ((w2 ^ 0xd800) < 0x400) {
                ssize_t l2 = mbrtowc(&w3, s + l, MB_CUR_MAX, &ps);
                if (l2 < 0) goto failed;
                if ((w3 ^ 0xdc00) >= 0x400) {
                    goto failed;
                }
                ch = ((w2 & 0x3ff) << 10) + (w3 & 0x3ff) + 0x10000;
                l += l2;
            } else ch = w2;
#else
            wchar_t w2;
            l = (ssize_t)mbrtowc(&w2, s, MB_CUR_MAX, &ps);
            if (l < 0) goto failed;
            ch = (unsigned int)w2;
#endif
        }
        i += l;
        if (j + 6 >= len) {
            char *d;
            len += 64;
            if (len < 64) goto failed;
            d = (char *)realloc(utf8, len);
            if (d == NULL) goto failed;
            utf8 = d;
        }
        if (ch < 0x80) utf8[j++] = (char)ch; else j += utf8out(ch, utf8 + j);
    }
    utf8[j] = 0;
    return utf8;
failed:
    free(utf8);
    return NULL;
}

int unc_path = 0;
int use_wide = 0;

wchar_t *utf8_to_wcharpath(const char *name) {
    size_t j = 0, len = strlen(name) + 2 + 4;
    wchar_t *wname = (wchar_t *)malloc(len * sizeof *wname);
    if (wname == NULL) return NULL;
    if (unc_path && (name[0] | 0x20) >= 'a' && (name[0] | 0x20) <= 'z' && name[1] == ':') {
        wname[0] = L'\\'; wname[1] = L'\\'; wname[2] = L'?'; wname[3] = L'\\';
        j = 4;
    }
    wname = utf8_to_wchar_common(name, wname, len, j);
    if (wname != NULL) {
        for (j = 0; wname[j] != L'\0'; j++) {
            if (wname[j] == L'/') wname[j] = L'\\';
        }
    }
    return wname;
}

char *utf8_to_charpath(const char *name) {
    size_t len = strlen(name) + 2 + 4;
    char *utf8 = (char *)malloc(len * sizeof *utf8);
    if (utf8 == NULL) return NULL;
    utf8 = utf8_to_char_common(name, utf8, len, 0);
    if (utf8 != NULL) {
        size_t j;
        for (j = 0; utf8[j] != '\0'; j++) {
            if (utf8[j] == '/') utf8[j] = '\\';
        }
    }
    return utf8;
}

int utf8_mkdir(const char *path) {
    int ret;
    if (use_wide) {
        wchar_t *wpath = utf8_to_wcharpath(path);
        if (wpath == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _wmkdir(wpath);
            free(wpath);
        }
    } else {
        char *path2 = utf8_to_charpath(path);
        if (path2 == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _mkdir(path2);
            free(path2);
        }
    }
    return ret;
}

int utf8_rmdir(const char *path) {
    int ret;
    if (use_wide) {
        wchar_t *wpath = utf8_to_wcharpath(path);
        if (wpath == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _wrmdir(wpath);
            free(wpath);
        }
    } else {
        char *path2 = utf8_to_charpath(path);
        if (path2 == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _rmdir(path2);
            free(path2);
        }
    }
    return ret;
}

int utf8_unlink(const char *path) {
    int ret;
    if (use_wide) {
        wchar_t *wpath = utf8_to_wcharpath(path);
        if (wpath == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _wunlink(wpath);
            free(wpath);
        }
    } else {
        char *path2 = utf8_to_charpath(path);
        if (path2 == NULL) {
            ret = -1;
            errno = ENOMEM;
        } else {
            ret = _unlink(path2);
            free(path2);
        }
    }
    return ret;
}

int utf8_open(const char *name, int flags, int mode) {
    int fd;
    if (use_wide) {
        wchar_t *wname = utf8_to_wcharpath(name);
        if (wname == NULL) {
            fd = -1;
            errno = ENOMEM;
        } else {
            fd = _wopen(wname, flags, mode);
            free(wname);
        }
    } else {
        char *path2 = utf8_to_charpath(name);
        if (path2 == NULL) {
            fd = -1;
            errno = ENOMEM;
        } else {
            fd = _open(path2, flags, mode);
            free(path2);
        }
    }
    return fd;
}
#endif
